// Package `tracking` defines a generic track (sort of record) type
// and a container type that allows storing of tracks in a SQL database.
package tracking

import (
	sqllib "database/sql"
	"encoding/json"
	"time"

	lib "git.sr.ht/~cco/go-scopes"
	"git.sr.ht/~cco/go-scopes/logging/log"
	"git.sr.ht/~cco/go-scopes/storage"
	sql "git.sr.ht/~cco/go-scopes/storage"
)

type ItemFactory func(*Container, ...string) Track

type Track interface {
	TrackId() lib.Ident
	Head() lib.StrMap
	Data() lib.Map
	TimeStamp() *time.Time
	SetTrackId(lib.Ident)
	SetTimeStamp(*time.Time)
	Update(lib.StrSlice, lib.Map, *time.Time)
}

type BaseTrack = track

// basic track implementation

type track struct {
	trackId   lib.Ident
	head      lib.StrMap
	timeStamp *time.Time
	data      lib.Map
	container *Container
}

func MakeTrack(cont *Container, h ...string) Track {
	head := lib.StrMap{}
	for i, k := range cont.HeadFields {
		if i >= len(h) {
			break
		}
		head[k] = h[i]
	}
	return &track{head: head}
}

func (tr *track) TrackId() lib.Ident {
	return tr.trackId
}

func (tr *track) Head() lib.StrMap {
	return tr.head
}

func (tr *track) Data() lib.Map {
	return tr.data
}

func (tr *track) TimeStamp() *time.Time {
	return tr.timeStamp
}

func (tr *track) SetTrackId(id lib.Ident) {
	tr.trackId = id
}

func (tr *track) SetTimeStamp(ts *time.Time) {
	tr.timeStamp = ts
}

func (tr *track) Update(h lib.StrSlice, d lib.Map, ts *time.Time) {
	for i, k := range tr.container.HeadFields {
		if i >= len(h) {
			break
		}
		if h[i] != "" {
			tr.head[k] = h[i]
		}
	}
	if d != nil {
		tr.data = d
	}
	if ts != nil {
		tr.timeStamp = ts
	}
}

// basic container implementation

type ContDef struct {
	ItemFactory ItemFactory
	TableName   string
	HeadFields  []string
	Indexes     [][]string
}

type Container struct {
	*ContDef
	Storage *sql.Storage
}

func Tracks(db *sql.Storage) *Container {
	return &Container{container_definition, db}
}

func (cont *Container) Get(id lib.Ident) Track {
	db := cont.Storage
	qu := lib.Map{
		"schema":    db.Schema,
		"tablename": cont.TableName,
		"scols":     append(cont.HeadFields, "timestamp", "data"),
		"qucols":    lib.StrSlice{"taskid"},
	}
	sql := storage.BuildSql(SqlSelect, qu)
	print(sql)
	var h []string
	tr := cont.ItemFactory(cont, h...)
	return tr
}

func (cont *Container) Query(headValues lib.StrSlice) Track {
	db := cont.Storage
	data := lib.Map{
		"schema":    db.Schema,
		"tablename": cont.TableName,
		"scols":     lib.StrSlice{"taskid", "data"},
		"qucols":    lib.StrSlice{"taskid", "username"},
		"ordcols":   lib.StrSlice{"timestamp"},
	}
	sql := storage.BuildSql(SqlSelect, data)
	//print(sql)
	_ = sql
	tr := cont.ItemFactory(cont)
	return tr
}

func (cont *Container) NewTrack(h []string, data lib.Map) Track {
	tr := cont.ItemFactory(cont, h...)
	//tr.SetData(data)
	cont.Insert(tr)
	return tr
}

func (cont *Container) Save(t Track) lib.Ident {
	return 0
}

func (cont *Container) Insert(tr Track) lib.Ident {
	var columns []string
	var values []any
	head := tr.Head()
	for _, k := range cont.HeadFields {
		columns = append(columns, k)
		values = append(values, head[k])
	}
	columns = append(columns, "Data")
	b, _ := json.Marshal(tr.Data())
	values = append(values, b)
	db := cont.Storage
	data := lib.Map{
		"schema":    db.Schema,
		"tablename": cont.TableName,
		"columns":   columns,
	}
	sql := storage.BuildSql(SqlInsert, data)
	var trid lib.Ident
	var ts *time.Time
	var tsstr string
	proc := func(r *sqllib.Rows) error {
		err := r.Scan(&trid, &ts)
		if err != nil {
			err = r.Scan(&trid, &tsstr)
			ts = ParseDateTime(tsstr)
		}
		return err
	}
	if err := db.Query(proc, sql, values...); err == nil {
		//tr.Update(&TrackTemplate{TrackId: trid, TimeStamp: ts})
		tr.SetTrackId(trid)
		tr.SetTimeStamp(ts)
		return trid
	}
	return 0
}

func ParseDateTime(inp string) *time.Time {
	ts, err := time.Parse("2006-01-02 15:04:05", inp)
	if err == nil {
		//fmt.Printf("%+v\n", ts)
		return &ts
	}
	log.Error(err).Msg("storage.tracking.ParseDateTime")
	return nil
}

func (cont *Container) CreateTable() {
	db := cont.Storage
	data := lib.Map{
		"schema":     db.Schema,
		"tablename":  cont.TableName,
		"headFields": cont.HeadFields,
		"indexes":    cont.Indexes,
		"params":     db.Params,
	}
	sql := storage.BuildSql(SqlCreate, data)
	if _, err := db.Exec(sql); err != nil {
		panic(err)
	}
}

// container definition

var container_definition *ContDef

func init() {
	container_definition = &ContDef{
		ItemFactory: MakeTrack,
		TableName:   "tracks",
		HeadFields:  []string{"taskId", "userName"},
		Indexes: [][]string{
			[]string{"taskId", "userName"},
			[]string{"userName"},
		},
	}
}
